# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
from enum import Enum

__all__ = (
    "ActionStatus",
    "StepStatus",
    "VersionType",
)


class VersionType(Enum):
    FULL = "FULL RELEASE"
    DEV = "DEV BUILD"
    RC = "RELEASE CANDIDATE"


class StepStatus(Enum):
    PASS = "PASS"
    FAIL = "FAIL"
    SKIP = "SKIP"


class ActionStatus(Enum):
    NOT_STARTED = "NOT STARTED"
    STARTED = "STARTED BUT NOT COMPLETED"
    COMPLETED = "COMPLETED"



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Generate visual representations of palettes in Bokeh palette groups.

The ``bokeh.palettes`` modules expose attributes such as ``mpl``, ``brewer``,
and ``d3`` that provide groups of palettes. The ``bokeh-palette-group``
directive accepts the name of one of these groups, and generates a visual
matrix of colors for every palette in the group.

As an example, the following usage of the the directive:

.. code-block:: rest

    .. bokeh-palette-group:: mpl

Generates the output:

    .. bokeh-palette-group:: mpl

.. note::
   This extension assumes both Bootstrap and JQuery are present (which is the
   case for the Bokeh documentation theme). If using this theme outside the
   Bokeh documentation, be sure to include those resources by hand.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# External imports
from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx.errors import SphinxError

# Bokeh imports
import bokeh.palettes as bp

# Bokeh imports
from .templates import PALETTE_GROUP_DETAIL

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'bokeh_palette_group',
    'BokehPaletteGroupDirective',
    'html_visit_bokeh_palette_group',
    'setup',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

class bokeh_palette_group(nodes.General, nodes.Element):
    pass


class BokehPaletteGroupDirective(Directive):

    has_content = False
    required_arguments = 1

    def run(self):
        node = bokeh_palette_group()
        node['group'] = self.arguments[0]
        return [node]

# NOTE: This extension now *assumes* both Bootstrap and JQuery are present
# (which is now the case for the Bokeh docs theme).
def html_visit_bokeh_palette_group(self, node):
    self.body.append('<div class="container-fluid"><div class="row">"')
    group = getattr(bp, node['group'], None)
    if not isinstance(group, dict):
        raise SphinxError("invalid palette group name %r" % node['group'])
    names = sorted(group)
    for name in names:
        palettes = group[name]
        # arbitrary cuttoff here, idea is to not show large (e.g 256 length) palettes
        numbers = [x for x in sorted(palettes) if x < 30]
        html = PALETTE_GROUP_DETAIL.render(name=name, numbers=numbers, palettes=palettes)
        self.body.append(html)
    self.body.append('</div></div>')
    raise nodes.SkipNode

def setup(app):
    ''' Required Sphinx extension setup function. '''
    app.add_node(bokeh_palette_group, html=(html_visit_bokeh_palette_group, None))
    app.add_directive('bokeh-palette-group', BokehPaletteGroupDirective)

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
import json
import re
from os.path import join
from subprocess import CalledProcessError

# External imports
from packaging.version import Version as V

# Bokeh imports
from .config import Config
from .system import cd, run
from .ui import failed, passed, skipped

__all__ = (
    "checkout_master",
    "checkout_release_branch",
    "commit",
    "delete_release_branch",
    "merge_release_branch",
    "push_to_github",
    "tag_version",
    "update_bokehjs_versions",
    "update_changelog",
    "update_hash_manifest",
)


def commit(config: Config, filename: str) -> None:
    path = join(config.repo_top_dir, filename)
    try:
        run(f"git add {path}")
    except CalledProcessError as e:
        failed("Could not git add {filename!r}", str(e).split("/n"))
        config.abort()
    try:
        run(f"git commit -m'Updating for version {config.version}'")
    except CalledProcessError as e:
        failed(f"Could not git commit {filename!r}", str(e).split("/n"))
        config.abort()
    passed(f"Committed file {filename!r}")


def checkout_release_branch(config: Config) -> None:
    try:
        run(f"git checkout -b {config.release_branch}")
        passed(f"Checked out release branch {config.release_branch!r}")
    except CalledProcessError as e:
        failed(f"Could not check out release branch {config.release_branch!r}", str(e).split("/n"))
        config.abort()


def checkout_master(config: Config) -> None:
    try:
        run("git checkout master")
        passed("Checked out master branch")
    except CalledProcessError as e:
        failed("Coud NOT check out master branch", str(e).split("/"))
        config.abort()


def merge_release_branch(config: Config) -> None:
    try:
        run(f"git merge --no-ff {config.release_branch} -m 'Merge branch {config.release_branch}'")
        passed("Merged release branch into master branch")
    except CalledProcessError as e:
        failed("Could NOT merge release branch in to master", str(e).split("/"))
        config.abort()


def delete_release_branch(config: Config) -> None:
    try:
        run(f"git branch -d {config.release_branch}")
        passed("Deleted release branch")
    except CalledProcessError as e:
        failed("Could NOT delete release branch", str(e).split("/"))
        config.abort()


def push_to_github(config: Config) -> None:
    try:
        # use --no-verify to prevent git hook that might ask for confirmation
        run("git push --no-verify origin master")
        passed("Pushed master branch to GitHub")
    except CalledProcessError as e:
        failed("Could NOT push master to origin", str(e).split("/"))
        config.abort()

    try:
        # use --no-verify to prevent git hook that might ask for confirmation
        run(f"git push --no-verify origin {config.version}")
        passed(f"Pushed tag {config.version!r} to GitHub")
    except CalledProcessError as e:
        failed("Could NOT push tag to origin", str(e).split("/"))
        config.abort()


def update_changelog(config: Config) -> None:
    if V(config.version).is_prerelease:
        skipped("Not updating CHANGELOG for pre-releases")
        return

    try:
        cd("scripts")
        run(f"python issues.py -p {config.last_full_version} -r {config.version}")
        cd("..")
        passed("Updated CHANGELOG with new closed issues")
        filename = join(config.repo_top_dir, "CHANGELOG")
        commit(config, filename)
    except CalledProcessError as e:
        if "HTTP Error 403: Forbidden" in e.output:
            failed("CHANGELOG cannot be updated right now due to GitHub rate limiting")
        else:
            failed("CHANGELOG update failed", e.output.split("\n"))


def update_bokehjs_versions(config: Config) -> None:

    regex_filenames = [
        "bokehjs/src/lib/version.ts",
    ]

    json_filenames = [
        "bokehjs/package.json",
        "bokehjs/package-lock.json",
    ]

    pat = rf"(release|version)([\" ][:=] [\"\']){config.js_last_any_version}([\"\'])"

    for filename in regex_filenames:
        path = join(config.repo_top_dir, filename)
        with open(path) as f:
            text = f.read()
            match = re.search(pat, text)

        if not match:
            failed(f"Unable to find version string for {config.js_last_any_version!r} in file {filename!r}")
            continue

        text = re.sub(pat, rf"\g<1>\g<2>{config.js_version}\g<3>", text)

        try:
            with open(path, "w") as f:
                f.write(text)
        except Exception as e:
            failed(f"Unable to write new version to file {filename!r}", str(e).split("\n"))
        else:
            passed(f"Updated version from {config.js_last_any_version!r} to {config.js_version!r} in file {filename!r}")
            commit(config, filename)

    for filename in json_filenames:
        path = join(config.repo_top_dir, filename)
        content = json.load(open(path))
        try:
            content["version"] = config.js_version
            with open(path, "w") as f:
                json.dump(content, f, indent=2)
                f.write("\n")
        except Exception as e:
            failed(f"Unable to write new version to file {filename!r}", str(e).split("\n"))
        else:
            passed(f"Updated version from {config.js_last_any_version!r} to {config.js_version!r} in file {filename!r}")
            commit(config, filename)


def update_hash_manifest(config: Config) -> None:
    if V(config.version).is_prerelease:
        skipped("Not updating SRH hash manifest for pre-releases")
        return

    try:
        cd("scripts")
        run(f"python sri.py {config.version}")
        cd("..")
        passed("Updated SRI hash manifest")
        commit(config, "bokeh/_sri.json")
    except CalledProcessError as e:
        failed("SRI hash manifest update failed", e.output.split("\n"))
        config.abort()


def tag_version(config: Config) -> None:
    try:
        run(f"git tag -a {config.version} -m 'Release {config.version}'", silent=False)
        passed("Tagged release {config.version!r}")
    except CalledProcessError as e:
        failed("COULD NOT TAG RELEASE" % e.output.split("\n"))
        config.abort()



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
import sys
from typing import Callable, Optional, Sequence

__all__ = (
    "banner",
    "failed",
    "passed",
    "skipped",
)

ColorFunction = Callable[[str], str]

try:
    import colorama  # isort:skip

    def bright(text: str) -> str:
        return "%s%s%s" % (colorama.Style.BRIGHT, text, colorama.Style.RESET_ALL)

    def dim(text: str) -> str:
        return "%s%s%s" % (colorama.Style.DIM, text, colorama.Style.RESET_ALL)

    def white(text: str) -> str:
        return "%s%s%s" % (colorama.Fore.WHITE, text, colorama.Style.RESET_ALL)

    def blue(text: str) -> str:
        return "%s%s%s" % (colorama.Fore.BLUE, text, colorama.Style.RESET_ALL)

    def red(text: str) -> str:
        return "%s%s%s" % (colorama.Fore.RED, text, colorama.Style.RESET_ALL)

    def green(text: str) -> str:
        return "%s%s%s" % (colorama.Fore.GREEN, text, colorama.Style.RESET_ALL)

    def yellow(text: str) -> str:
        return "%s%s%s" % (colorama.Fore.YELLOW, text, colorama.Style.RESET_ALL)

    sys.platform == "win32" and colorama.init()
except ImportError:

    def bright(text: str) -> str:
        return text

    def dim(text: str) -> str:
        return text

    def white(text: str) -> str:
        return text

    def blue(text: str) -> str:
        return text

    def red(text: str) -> str:
        return text

    def green(text: str) -> str:
        return text

    def yellow(text: str) -> str:
        return text


def banner(color: ColorFunction, msg: str) -> None:
    print()
    print(color("=" * 80))
    print(color("{:^80}".format(msg)))
    print(color("=" * 80 + "\n"))


def failed(msg: str, details: Optional[Sequence[str]] = None) -> None:
    print((red("[FAIL] ")) + msg)
    if details:
        print()
        for line in details:
            print("     " + line)
        print()


def passed(msg: str) -> None:
    print(dim(green("[PASS] ")) + msg)


def skipped(msg: str) -> None:
    print(blue("[SKIP] ") + msg)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
from typing import Tuple

# Bokeh imports
from .build import (
    build_bokehjs,
    build_conda_packages,
    build_docs,
    build_sdist_packages,
    dev_install,
    install_bokehjs,
    npm_install,
)
from .checks import (
    check_checkout,
    check_git,
    check_issues,
    check_release_branch,
    check_repo,
    check_tags,
    check_version_order,
)
from .config import StepType
from .credentials import (
    verify_anaconda_credentials,
    verify_aws_credentials,
    verify_npm_credentials,
    verify_pypi_credentials,
)
from .git import (
    checkout_master,
    checkout_release_branch,
    delete_release_branch,
    merge_release_branch,
    push_to_github,
    tag_version,
    update_bokehjs_versions,
    update_changelog,
    update_hash_manifest,
)
from .publish import (
    publish_anaconda,
    publish_cdn,
    publish_docs,
    publish_npm,
    publish_pypi,
)

__all__ = (
    "BUILD_CHECKS",
    "BUILD_STEPS",
    "DEPLOY_CHECKS",
    "DEPLOY_STEPS",
)

StepListType = Tuple[StepType, ...]

BUILD_CHECKS: StepListType = (
    check_git,
    check_repo,
    check_checkout,
    check_tags,
    check_version_order,
    check_release_branch,
    check_issues,
    verify_aws_credentials,
)

BUILD_STEPS: StepListType = (
    # clean_repo,
    checkout_release_branch,
    update_bokehjs_versions,
    update_changelog,
    npm_install,
    build_bokehjs,
    install_bokehjs,
    dev_install,
    update_hash_manifest,
    tag_version,
    build_conda_packages,
    build_sdist_packages,
    build_docs,
    # pack_deployment,
    # upload_deployment,
    publish_cdn,
    checkout_master,
    merge_release_branch,
    push_to_github,
    delete_release_branch,
)

DEPLOY_CHECKS: StepListType = (
    # check_aws_cli,
    verify_anaconda_credentials,
    verify_pypi_credentials,
    verify_npm_credentials,
    verify_aws_credentials,
)

DEPLOY_STEPS: StepListType = (
    # download_deployment,
    # unpack_deployment,
    publish_docs,
    # update_versions_config,
    publish_anaconda,
    publish_pypi,
    publish_npm,
    # issue_cdn_invalidations,
)
